/*****************************************************************************/
/*
  animate.c: animate a data file.
*/
/*****************************************************************************/
/* This version flips X and angles to get lighting right */
#define XFLIP -
#define XXFLIP 

/*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#ifdef WIN32
#include <windows.h>
#else
#include <time.h>
#endif

#include "drawstuff.h" /* Graphics stuff */
#include "drawstuff-cga.h" /* CGA stuff to make things clearer */

#include "useful/useful.h"
#include "mrdplot/mrdplot.h"
#include "dynamics.h"

#include "data.h"

/*****************************************************************************/

/* These are handy constants. XX rather than X to avoid math.h conflicts */
#define XX 0
#define YY 1
#define ZZ 2

/*****************************************************************************/

Dynamics *d = NULL;

MRDPLOT_DATA *mrdplot_d = NULL;

int channels[N_CHANNELS];

double the_start_time = 0;
double start_time = 0;
double data_start_time = 0;

/*****************************************************************************/
/************************************************************************/

#ifdef WIN32
double get_time()
{
  return (double) timeGetTime()/1000.0;
}
#else
double get_time()
{
  struct timespec the_tp;

  clock_gettime( CLOCK_MONOTONIC, &the_tp );
  return ((double) (the_tp.tv_sec)) + 1.0e-9*the_tp.tv_nsec;
}
#endif

/*****************************************************************************/

double lookup_value( MRDPLOT_DATA *d, int point, int channel )
{
  if ( d == NULL )
    {
      fprintf( stderr, "No MRDPLOT data.\n" );
      exit( -1 );
    }
  if ( channels[channel] < 0 || channels[channel] >= d->n_channels )
    {
      fprintf( stderr, "Bad channel: %d %d\n", channel, channels[channel] );
      exit( -1 );
    }
  return d->data[point*mrdplot_d->n_channels + channels[channel]];
}

/********************************************************************/
// start(): set things up

static void start()
{
  // set up view point
  static float xyz[3] = {1.0f,0.0f,0.14f};
  // static float hpr[3] = {-180.0000f,0.0000f,0.0000f};
  static float hpr[3] = {180.0f,27.0f,0.0f};
  dsSetViewpoint (xyz,hpr);
}

/********************************************************************/
// called when a key pressed

static void command (int cmd)
{
  // don't handle user input yet.
  dsPrint ("received command %d (`%c')\n",cmd,cmd);
}

/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/

#define ROLLER_WIDTH 0.2
#define BAR_WIDTH 0.02
#define MARKER_SIZE 0.03

static void display (int pause)
{
  static int index = 0;
  float f;
  float center[3];
  float R_identity[12];
  float R_roller[12];
  float R_board[12];
  float R_shank[12];
  float R_thigh[12];
  float R_torso[12];
  int i;
  float board_sides[3];
  float shank_sides[3];
  float thigh_sides[3];
  float torso_sides[3];
  float marker[3];
  float data_point_time;
  double the_time;
  double xr, xc, yc, xb, yb;
  double xal, yal, xsl, ysl, x_shankl, y_shankl, xkl, ykl, xthl, ythl, x_thighl, y_thighl;
  double xar, yar, xsr, ysr, x_shankr, y_shankr, xkr, ykr, xthr, ythr, x_thighr, y_thighr;
  double xh, yh, xto, yto, x_torso, y_torso;
  double xhead, yhead;
  double ab, cb, sb, ato, cato, sato;
  double asl, casl, sasl, athl, cathl, sathl;
  double asr, casr, sasr, athr, cathr, sathr;
  double xcom, ycom;

  xr = lookup_value( mrdplot_d, index, M_XR );
  xc = lookup_value( mrdplot_d, index, M_XC );
  yc = lookup_value( mrdplot_d, index, M_YC );
  xb = lookup_value( mrdplot_d, index, M_XB );
  yb = lookup_value( mrdplot_d, index, M_YB );
  xal = lookup_value( mrdplot_d, index, M_XAL );
  yal = lookup_value( mrdplot_d, index, M_YAL );
  xsl = lookup_value( mrdplot_d, index, M_XSL );
  ysl = lookup_value( mrdplot_d, index, M_YSL );
  xkl = lookup_value( mrdplot_d, index, M_XKL );
  ykl = lookup_value( mrdplot_d, index, M_YKL );
  xthl = lookup_value( mrdplot_d, index, M_XTHL );
  ythl = lookup_value( mrdplot_d, index, M_YTHL );
  xar = lookup_value( mrdplot_d, index, M_XAR );
  yar = lookup_value( mrdplot_d, index, M_YAR );
  xsr = lookup_value( mrdplot_d, index, M_XSR );
  ysr = lookup_value( mrdplot_d, index, M_YSR );
  xkr = lookup_value( mrdplot_d, index, M_XKR );
  ykr = lookup_value( mrdplot_d, index, M_YKR );
  xthr = lookup_value( mrdplot_d, index, M_XTHR );
  ythr = lookup_value( mrdplot_d, index, M_YTHR );
  xh = lookup_value( mrdplot_d, index, M_XH );
  yh = lookup_value( mrdplot_d, index, M_YH );
  xto = lookup_value( mrdplot_d, index, M_XTO );
  yto = lookup_value( mrdplot_d, index, M_YTO );
  xhead = lookup_value( mrdplot_d, index, M_XHEAD );
  yhead = lookup_value( mrdplot_d, index, M_YHEAD );
  xcom = lookup_value( mrdplot_d, index, M_XCOM );
  ycom = lookup_value( mrdplot_d, index, M_YCOM );
  x_shankl = (xal + xkl)/2;
  y_shankl = (yal + ykl)/2;
  x_thighl = (xkl + xh)/2;
  y_thighl = (ykl + yh)/2;
  x_shankr = (xar + xkr)/2;
  y_shankr = (yar + ykr)/2;
  x_thighr = (xkr + xh)/2;
  y_thighr = (ykr + yh)/2;
  x_torso = (xh + xhead)/2;
  y_torso = (yh + yhead)/2;
  ab = lookup_value( mrdplot_d, index, M_AB );
  asl = lookup_value( mrdplot_d, index, M_ASL );
  athl = lookup_value( mrdplot_d, index, M_ATHL );
  asr = lookup_value( mrdplot_d, index, M_ASR );
  athr = lookup_value( mrdplot_d, index, M_ATHR );
  ato = lookup_value( mrdplot_d, index, M_ATO );
  cb = cos( ab );
  sb = sin( ab );
  casl = cos( asl );
  sasl = sin( asl );
  cathl = cos( athl );
  sathl = sin( athl );
  casr = cos( asr );
  sasr = sin( asr );
  cathr = cos( athr );
  sathr = sin( athr );
  cato = cos( ato );
  sato = sin( ato );

  // set up identity matrix
  for ( i = 0; i < 12; i++ )
    R_identity[i] = 0;
  R_identity[0] = R_identity[5] = R_identity[10] = 1.0;

  /* Roller */
  center[XX] = 0;
  center[YY] = xr;
  center[ZZ] = d->roller_radius;
  for ( i = 0; i < 12; i++ )
    R_roller[i] = 0;
  R_roller[2] = R_roller[5] = R_roller[8] = 1.0;
  dsSetTexture (DS_NONE);
  dsSetColor (1,0,0);
  dsDrawCylinder ( center, R_roller, (float) ROLLER_WIDTH,
		   (float) (d->roller_radius) );

  /* Board */
  center[XX] = 0;
  center[YY] = xb;
  center[ZZ] = yb;
  board_sides[XX] = ROLLER_WIDTH;
  board_sides[YY] = d->board_length;
  board_sides[ZZ] = BAR_WIDTH;
  for ( i = 0; i < 12; i++ )
    R_board[i] = 0;
  R_board[0] = 1.0;
  R_board[5] = cb;
  R_board[6] = -sb;
  R_board[9] = sb;
  R_board[10] = cb;
  dsSetTexture (DS_NONE);
  dsSetColor (0,1,0);
  dsDrawBox( center, R_board, board_sides );

  /* Left Shank */
  center[XX] = 0;
  center[YY] = x_shankl;
  center[ZZ] = y_shankl;
  shank_sides[XX] = BAR_WIDTH;
  shank_sides[YY] = BAR_WIDTH;
  shank_sides[ZZ] = d->shank_length;
  for ( i = 0; i < 12; i++ )
    R_shank[i] = 0;
  R_shank[0] = 1.0;
  R_shank[5] = casl;
  R_shank[6] = -sasl;
  R_shank[9] = sasl;
  R_shank[10] = casl;
  dsSetTexture (DS_NONE);
  dsSetColor (0,0,1);
  dsDrawBox( center, R_shank, shank_sides );

  /* Left Thigh */
  center[XX] = 0;
  center[YY] = x_thighl;
  center[ZZ] = y_thighl;
  thigh_sides[XX] = BAR_WIDTH;
  thigh_sides[YY] = BAR_WIDTH;
  thigh_sides[ZZ] = d->thigh_length;
  for ( i = 0; i < 12; i++ )
    R_thigh[i] = 0;
  R_thigh[0] = 1.0;
  R_thigh[5] = cathl;
  R_thigh[6] = -sathl;
  R_thigh[9] = sathl;
  R_thigh[10] = cathl;
  dsSetTexture (DS_NONE);
  dsSetColor (0,0,1);
  dsDrawBox( center, R_thigh, thigh_sides );

  /* Right Shank */
  center[XX] = 0;
  center[YY] = x_shankr;
  center[ZZ] = y_shankr;
  shank_sides[XX] = BAR_WIDTH;
  shank_sides[YY] = BAR_WIDTH;
  shank_sides[ZZ] = d->shank_length;
  for ( i = 0; i < 12; i++ )
    R_shank[i] = 0;
  R_shank[0] = 1.0;
  R_shank[5] = casr;
  R_shank[6] = -sasr;
  R_shank[9] = sasr;
  R_shank[10] = casr;
  dsSetTexture (DS_NONE);
  dsSetColor (0,0,1);
  dsDrawBox( center, R_shank, shank_sides );

  /* Right Thigh */
  center[XX] = 0;
  center[YY] = x_thighr;
  center[ZZ] = y_thighr;
  thigh_sides[XX] = BAR_WIDTH;
  thigh_sides[YY] = BAR_WIDTH;
  thigh_sides[ZZ] = d->thigh_length;
  for ( i = 0; i < 12; i++ )
    R_thigh[i] = 0;
  R_thigh[0] = 1.0;
  R_thigh[5] = cathr;
  R_thigh[6] = -sathr;
  R_thigh[9] = sathr;
  R_thigh[10] = cathr;
  dsSetTexture (DS_NONE);
  dsSetColor (0,0,1);
  dsDrawBox( center, R_thigh, thigh_sides );

  /* Torso */
  center[XX] = 0;
  center[YY] = x_torso;
  center[ZZ] = y_torso;
  torso_sides[XX] = BAR_WIDTH;
  torso_sides[YY] = BAR_WIDTH;
  torso_sides[ZZ] = d->torso_length;
  for ( i = 0; i < 12; i++ )
    R_torso[i] = 0;
  R_torso[0] = 1.0;
  R_torso[5] = cato;
  R_torso[6] = -sato;
  R_torso[9] = sato;
  R_torso[10] = cato;
  dsSetTexture (DS_NONE);
  dsSetColor (0,0,1);
  dsDrawBox( center, R_torso, torso_sides );

  /* Markers */
  marker[XX] = 0.0;
  marker[YY] = xc;
  marker[ZZ] = yc;
  dsSetTexture (DS_NONE);
  dsSetColor (1,1,1);
  dsDrawSphere( marker, R_identity, MARKER_SIZE );

  marker[XX] = 0.0;
  marker[YY] = xb;
  marker[ZZ] = yb;
  dsSetTexture (DS_NONE);
  dsSetColor (1,1,1);
  dsDrawSphere( marker, R_identity, MARKER_SIZE );

  marker[XX] = 0.0;
  marker[YY] = xal;
  marker[ZZ] = yal;
  dsSetTexture (DS_NONE);
  dsSetColor (1,1,1);
  dsDrawSphere( marker, R_identity, MARKER_SIZE );

  marker[XX] = 0.0;
  marker[YY] = xsl;
  marker[ZZ] = ysl;
  dsSetTexture (DS_NONE);
  dsSetColor (1,1,1);
  dsDrawSphere( marker, R_identity, MARKER_SIZE );

  marker[XX] = 0.0;
  marker[YY] = xkl;
  marker[ZZ] = ykl;
  dsSetTexture (DS_NONE);
  dsSetColor (1,1,1);
  dsDrawSphere( marker, R_identity, MARKER_SIZE );

  marker[XX] = 0.0;
  marker[YY] = xthl;
  marker[ZZ] = ythl;
  dsSetTexture (DS_NONE);
  dsSetColor (1,1,1);
  dsDrawSphere( marker, R_identity, MARKER_SIZE );

  marker[XX] = 0.0;
  marker[YY] = xar;
  marker[ZZ] = yar;
  dsSetTexture (DS_NONE);
  dsSetColor (1,1,1);
  dsDrawSphere( marker, R_identity, MARKER_SIZE );

  marker[XX] = 0.0;
  marker[YY] = xsr;
  marker[ZZ] = ysr;
  dsSetTexture (DS_NONE);
  dsSetColor (1,1,1);
  dsDrawSphere( marker, R_identity, MARKER_SIZE );

  marker[XX] = 0.0;
  marker[YY] = xkr;
  marker[ZZ] = ykr;
  dsSetTexture (DS_NONE);
  dsSetColor (1,1,1);
  dsDrawSphere( marker, R_identity, MARKER_SIZE );

  marker[XX] = 0.0;
  marker[YY] = xthr;
  marker[ZZ] = ythr;
  dsSetTexture (DS_NONE);
  dsSetColor (1,1,1);
  dsDrawSphere( marker, R_identity, MARKER_SIZE );

  marker[XX] = 0.0;
  marker[YY] = xh;
  marker[ZZ] = yh;
  dsSetTexture (DS_NONE);
  dsSetColor (1,1,1);
  dsDrawSphere( marker, R_identity, MARKER_SIZE );

  marker[XX] = 0.0;
  marker[YY] = xto;
  marker[ZZ] = yto;
  dsSetTexture (DS_NONE);
  dsSetColor (1,1,1);
  dsDrawSphere( marker, R_identity, MARKER_SIZE );

  marker[XX] = 0.0;
  marker[YY] = xhead;
  marker[ZZ] = yhead;
  dsSetTexture (DS_NONE);
  dsSetColor (1,1,1);
  dsDrawSphere( marker, R_identity, MARKER_SIZE );

  marker[XX] = 0.0;
  marker[YY] = xcom;
  marker[ZZ] = ycom;
  dsSetTexture (DS_NONE);
  dsSetColor (1,0,0);
  dsDrawSphere( marker, R_identity, MARKER_SIZE );

  data_point_time = lookup_value( mrdplot_d, index, M_TIME );

  /* Delay */
  for( ; ; )
    {
      the_time = get_time();
      if ( the_time - start_time > data_point_time - data_start_time )
	break;
#ifdef WIN32
      for( f = 0.0; f < 10000.0; f += 1.0 )
	;
#else
      for( f = 0.0; f < 10000.0; f += 1.0 )
	;
#endif
    }

  index += 16;
  if ( index >= mrdplot_d->n_points )
    {
      index = 0;
      data_start_time = lookup_value( mrdplot_d, 0, M_TIME );
      start_time = get_time();
      the_time = get_time();
      printf( "time: %g\n", the_time - the_start_time );
      start_time = the_time;
    }
}

/*****************************************************************************/

void read_data_file(int argc, char **argv)
{
  char *filename;
  int i;


  if ( argc > 1 )
    filename = strdup( argv[1] );
  else
    filename = last_data();
  
  printf( "Reading: %s\n", filename );

  mrdplot_d = read_mrdplot( filename );

  for( i = 0; i < N_CHANNELS; i++ )
    channels[i] = -1;

  channels[M_TIME] = find_channel( "time", mrdplot_d );
  channels[M_XR] = find_channel( "xr", mrdplot_d );
  channels[M_XC] = find_channel( "xc", mrdplot_d );
  channels[M_YC] = find_channel( "yc", mrdplot_d );
  channels[M_XB] = find_channel( "xb", mrdplot_d );
  channels[M_YB] = find_channel( "yb", mrdplot_d );
  channels[M_AB] = find_channel( "ab", mrdplot_d );
  channels[M_XAL] = find_channel( "xal", mrdplot_d );
  channels[M_YAL] = find_channel( "yal", mrdplot_d );
  channels[M_XSL] = find_channel( "xsl", mrdplot_d );
  channels[M_YSL] = find_channel( "ysl", mrdplot_d );
  channels[M_ASL] = find_channel( "asl", mrdplot_d );
  channels[M_XKL] = find_channel( "xkl", mrdplot_d );
  channels[M_YKL] = find_channel( "ykl", mrdplot_d );
  channels[M_XTHL] = find_channel( "xthl", mrdplot_d );
  channels[M_YTHL] = find_channel( "ythl", mrdplot_d );
  channels[M_ATHL] = find_channel( "athl", mrdplot_d );
  channels[M_XAR] = find_channel( "xar", mrdplot_d );
  channels[M_YAR] = find_channel( "yar", mrdplot_d );
  channels[M_XSR] = find_channel( "xsr", mrdplot_d );
  channels[M_YSR] = find_channel( "ysr", mrdplot_d );
  channels[M_ASR] = find_channel( "asr", mrdplot_d );
  channels[M_XKR] = find_channel( "xkr", mrdplot_d );
  channels[M_YKR] = find_channel( "ykr", mrdplot_d );
  channels[M_XTHR] = find_channel( "xthr", mrdplot_d );
  channels[M_YTHR] = find_channel( "ythr", mrdplot_d );
  channels[M_ATHR] = find_channel( "athr", mrdplot_d );
  channels[M_XH] = find_channel( "xh", mrdplot_d );
  channels[M_YH] = find_channel( "yh", mrdplot_d );
  channels[M_XTO] = find_channel( "xto", mrdplot_d );
  channels[M_YTO] = find_channel( "yto", mrdplot_d );
  channels[M_ATO] = find_channel( "ato", mrdplot_d );
  channels[M_XHEAD] = find_channel( "xhead", mrdplot_d );
  channels[M_YHEAD] = find_channel( "yhead", mrdplot_d );
  channels[M_XCOM] = find_channel( "xcom", mrdplot_d );
  channels[M_YCOM] = find_channel( "ycom", mrdplot_d );

  for( i = 0; i < N_CHANNELS; i++ )
    printf( "%d %d\n", i, channels[i] );
}

/************************************************************************/
/********************************************************************/

int main (int argc, char **argv)
{
  dsFunctions fn; 

  init_dynamics();
  d = create_dynamics();

  read_data_file( argc, argv );

  // setup pointers to drawstuff callback functions
  fn.version = DS_VERSION; 
  fn.start = &start; 
  fn.step = &display; 
  fn.command = &command; 
  fn.stop = 0; 
#ifdef WIN32
  fn.path_to_textures = "../drawstuff-windows/textures";
#else
  fn.path_to_textures = "../drawstuff-linux/textures";
#endif

  the_start_time = start_time = get_time();

  data_start_time = lookup_value( mrdplot_d, 0, M_TIME );

  // do display
  dsSimulationLoop( argc, argv, /* command line arguments */
                    500, 500, /* window size */
                    &fn ); /* callback info */

  return 0;
}

/********************************************************************/
