/************************************************************************/
/*
dynamics for bongo-boarding (two legs with knees)
*/
/***********************************************************************/
/* Notes:
1) accelerations are stored with old action
2) since set_state does not have an action, the accelerations are not
   meaningful after set_state().

/************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "useful/useful.h"
#include "nr2d/nrutil.h"
#include "dm/dm.h"
#include "dynamics.h"

/**********************************************************************/
/**********************************************************************/
// Mathematica vesions of function calls

double Sin( double x )
{
  return sin( x );
}

double Cos( double x )
{
  return cos( x );
}

// CGA: handle special case of integer y more efficiently
double Power( double x, double y )
{
  if ( y == 2.0 )
    return x*x;
  /*
  if ( y != 2.0 )
    {
      fprintf( stderr, "Power( %g, %g )\n", x, y );
      exit( -1 );
    }
  */
  return pow( x, y );
}

double Sqrt( double x )
{
  return sqrt( x );
}

double Csc( double x )
{
  return 1.0/sin( x );
}

double ArcCos( double x )
{
  return acos( x );
}

/**********************************************************************/
/**********************************************************************/
/* Don't need to do any initialization */

int init_dynamics()
{
  return 1;
}

/**********************************************************************/

void forward_kinematics( Dynamics *d )
{
  double xr0, mr, Ir, rr, mb, Ib, scm, sl, ms, Is, thcm, thl, mth, Ith;
  double tocm, tol, mto, Ito, ao;
  double abd, asld, asrd, athld, athrd, atod;
  double sab, cab, sasl, casl, sasr, casr, sathl, cathl, sathr, cathr;
  double sato, cato;

  // use Mathematica variable names
  xr0 = d->roller_x0;
  rr = d->roller_radius;
  mr = d->roller_mass;
  Ir = d->roller_I;
  mb = d->board_mass;
  Ib = d->board_I;
  scm = d->shank_cm;
  sl = d->shank_length;
  ms = d->shank_mass;
  Is = d->shank_I;
  thcm = d->thigh_cm;
  thl = d->thigh_length;
  mth = d->thigh_mass;
  Ith = d->thigh_I;
  tocm = d->torso_cm;
  tol = d->torso_length;
  mto = d->torso_mass;
  Ito = d->torso_I;
  ao = d->ao;

  // use Mathematica variable names
  d->ar = -d->full_state[XX_XR]/rr;
  d->ab = d->full_state[XX_AB];
  d->asl = d->full_state[XX_ASL];
  d->asr = d->full_state[XX_ASR];
  d->athl = d->full_state[XX_ATHL];
  d->athr = d->full_state[XX_ATHR];
  d->ato = d->full_state[XX_ATO];

  d->xrd = d->full_state[XX_XR + N_XX_ACC];
  d->wr = -(d->xrd/rr);
  abd = d->wb = d->full_state[XX_AB + N_XX_ACC];
  asld = d->wsl = d->full_state[XX_ASL + N_XX_ACC];
  asrd = d->wsr = d->full_state[XX_ASR + N_XX_ACC];
  athld = d->wthl = d->full_state[XX_ATHL + N_XX_ACC];
  athrd = d->wthr = d->full_state[XX_ATHR + N_XX_ACC];
  atod = d->wto = d->full_state[XX_ATO + N_XX_ACC];

  // do all the sin/cos calls
  sab = sin( d->ab );
  cab = cos( d->ab );
  sasl = sin( d->asl );
  casl = cos( d->asl );
  sasr = sin( d->asr );
  casr = cos( d->asr );
  sathl = sin( d->athl );
  cathl = cos( d->athl );
  sathr = sin( d->athr );
  cathr = cos( d->athr );
  sato = sin( d->ato );
  cato = cos( d->ato );

  // use Mathematica variable names
  /* We are switching back to d-> versions
  xb = d->full_state[XX_XB];
  yb = d->full_state[XX_YB];
  xsl = d->full_state[XX_XSL];
  ysl = d->full_state[XX_YSL];
  xthl = d->full_state[XX_XTHL];
  ythl = d->full_state[XX_YTHL];
  xsr = d->full_state[XX_XSR];
  ysr = d->full_state[XX_YSR];
  xthr = d->full_state[XX_XTHR];
  ythr = d->full_state[XX_YTHR];
  xto = d->full_state[XX_XTO];
  yto = d->full_state[XX_YTO];
  */

  // use Mathematica variable names
  /* We are switching back to d-> versions
  xbd = d->full_state[XX_XB + N_XX_ACC];
  ybd = d->full_state[XX_YB + N_XX_ACC];
  xsld = d->full_state[XX_XSL + N_XX_ACC];
  ysld = d->full_state[XX_YSL + N_XX_ACC];
  xthld = d->full_state[XX_XTHL + N_XX_ACC];
  ythld = d->full_state[XX_YTHL + N_XX_ACC];
  xsrd = d->full_state[XX_XSR + N_XX_ACC];
  ysrd = d->full_state[XX_YSR + N_XX_ACC];
  xthrd = d->full_state[XX_XTHR + N_XX_ACC];
  ythrd = d->full_state[XX_YTHR + N_XX_ACC];
  xtod = d->full_state[XX_XTO + N_XX_ACC];
  ytod = d->full_state[XX_YTO + N_XX_ACC];
  */

  // As much as possible the following is raw Mathematica output

  // positions

  d->xr = d->full_state[XX_XR];
  d->yr = rr;

  d->xc = d->xr +
    -(rr*sab);

  d->yc = d->yr +
    rr*cab;

  d->xb = d->xc +
    (d->ab*rr + d->xr - xr0)*cab;

  d->yb = d->yc +
    (d->ab*rr + d->xr - xr0)*sab;

  d->xal = d->xb +
    ao*cab;

  d->yal = d->yb +
    ao*sab;

  d->xsl = d->xal +
    -(scm*sasl);

  d->ysl = d->yal +
    scm*casl;

  d->xkl = d->xal +
    -(sl*sasl);

  d->ykl = d->yal +
    sl*casl;

  d->xthl = d->xkl +
    -(thcm*sathl);

  d->ythl = d->ykl +
    thcm*cathl;

  d->xar = d->xb + 
    -(ao*cab);

  d->yar = d->yb +
    -(ao*sab);

  d->xsr = d->xar +
    -(scm*sasr);

  d->ysr = d->yar +
    scm*casr;

  d->xkr = d->xar +
    -(sl*sasr);

  d->ykr = d->yar +
    sl*casr;

  d->xthr = d->xkr +
    -(thcm*sathr);

  d->ythr = d->ykr +
    thcm*cathr;

  d->xh = d->xkl +
    -(thl*sathl);

  d->yh = d->ykl +
    thl*cathl;

  d->xto = d->xh +
    -(tocm*sato);

  d->yto = d->yh +
    tocm*cato;

  d->xhead = d->xh +
    -(tol*sato);

  d->yhead = d->yh +
    tol*cato;

  // velocities

  // xrd already set above with angles
  d->yrd = 0.0;

  d->xcd = d->xrd +
    -(abd*rr*cab);

  d->ycd = d->yrd +
    -(abd*rr*sab);

  d->xbd = d->xcd +
    (abd*rr + d->xrd)*cab - abd*(d->ab*rr + d->xr - xr0)*sab;

  d->ybd = d->ycd +
    abd*(d->ab*rr + d->xr - xr0)*cab + (abd*rr + d->xrd)*sab;

  d->xald = d->xbd +
    -(abd*ao*sab);

  d->yald = d->ybd +
    abd*ao*cab;

  d->xsld = d->xald +
    -(asld*scm*casl);

  d->ysld = d->yald +
    -(asld*scm*sasl);

  d->xkld = d->xald +
    -(asld*sl*casl);

  d->ykld = d->yald +
    -(asld*sl*sasl);

  d->xthld = d->xkld +
    -(athld*thcm*cathl);

  d->ythld = d->ykld +
    -(athld*thcm*sathl);

  d->xard = d->xbd +
    (abd*ao*sab);

  d->yard = d->ybd +
    -abd*ao*cab;

  d->xsrd = d->xard +
    -(asrd*scm*casr);

  d->ysrd = d->yard +
    -(asrd*scm*sasr);

  d->xkrd = d->xard +
    -(asrd*sl*casr);

  d->ykrd = d->yard +
    -(asrd*sl*sasr);

  d->xthrd = d->xkrd +
    -(athrd*thcm*cathr);

  d->ythrd = d->ykrd +
    -(athrd*thcm*sathr);

  d->xhd = d->xkld +
    -(athld*thl*cathl);

  d->yhd = d->ykld +
    -(athld*thl*sathl);

  d->xtod = d->xhd +
    -(atod*tocm*cato);

  d->ytod = d->yhd +
    -(atod*tocm*sato);

  d->xheadd = d->xhd +
    -(atod*tol*cato);

  d->yheadd = d->yhd +
    -(atod*tol*sato);

  /*
  d->wb =
    abd;
  d->wsl =
    asld;
  d->wthl =
    athld;
  d->wsr =
    asrd;
  d->wthr =
    athrd;
  d->wto =
    atod;
  */

  // energies

  d->rke = d->wr*Ir*d->wr/2 + mr*d->xrd*d->xrd/2;
  d->rpe = mr*G*rr;

  d->bke = d->wb*Ib*d->wb/2 + mb*d->xbd*d->xbd/2 + mb*d->ybd*d->ybd/2;
  d->bpe = mb*G*d->yb;

  d->slke = d->wsl*Is*d->wsl/2 + ms*d->xsld*d->xsld/2 + ms*d->ysld*d->ysld/2;
  d->slpe = ms*G*d->ysl;

  d->thlke = d->wthl*Ith*d->wthl/2 + mth*d->xthld*d->xthld/2 + mth*d->ythld*d->ythld/2;
  d->thlpe = mth*G*d->ythl;

  d->srke = d->wsr*Is*d->wsr/2 + ms*d->xsrd*d->xsrd/2 + ms*d->ysrd*d->ysrd/2;
  d->srpe = ms*G*d->ysr;

  d->thrke = d->wthr*Ith*d->wthr/2 + mth*d->xthrd*d->xthrd/2 + mth*d->ythrd*d->ythrd/2;
  d->thrpe = mth*G*d->ythr;

  d->toke = d->wto*Ito*d->wto/2 + mto*d->xtod*d->xtod/2 + mto*d->ytod*d->ytod/2;
  d->tope = mto*G*d->yto;

  d->ke = d->rke + d->bke + d->slke + d->thlke + d->srke + d->thrke + d->toke;
  d->pe = d->rpe + d->bpe + d->slpe + d->thlpe + d->srpe + d->thrpe + d->tope;

  d->energy = d->ke + d->pe;

  d->xcom = (mr*d->xr + mb*d->xb + ms*d->xsl + mth*d->xthl + ms*d->xsr + mth*d->xthr + mto*d->xto)/d->total_mass;
  d->ycom = (mr*rr + mb*d->yb + ms*d->ysl + mth*d->ythl + ms*d->ysr + mth*d->ythr + mto*d->yto)/d->total_mass;
  d->dx = d->xcom - d->xr;

  d->xcomd = (mr*d->xrd + mb*d->xbd + ms*d->xsld + mth*d->xthld + ms*d->xsrd + mth*d->xthrd + mto*d->xtod)/d->total_mass;
  d->ycomd = (mb*d->ybd + ms*d->ysld + mth*d->ythld + ms*d->ysrd + mth*d->ythrd + mto*d->ytod)/d->total_mass;
  d->dxd = d->xcomd - d->xrd;
}

/**********************************************************************/

void full_state_to_state( Dynamics *d, double *state )
{
  double sl, thl;
  double ao;
  double asl, athl, asr, athr;
  double asld, athld, asrd, athrd;
  double abd;
  double pi = 3.14159265;
  double akl, akr, akld, akrd;
  double errorl, errorr;
  double lll2, llr2;

  sl = d->shank_length;
  thl = d->thigh_length;
  ao = d->ao;

  asl = d->full_state[XX_ASL];
  athl = d->full_state[XX_ATHL];
  asr = d->full_state[XX_ASR];
  athr = d->full_state[XX_ATHR];

  abd = d->full_state[XX_AB + N_XX_ACC];

  asld = d->full_state[XX_ASL + N_XX_ACC];
  athld = d->full_state[XX_ATHL + N_XX_ACC];
  asrd = d->full_state[XX_ASR + N_XX_ACC];
  athrd = d->full_state[XX_ATHR + N_XX_ACC];

  state[X_XR] = d->full_state[XX_XR];
  state[X_AB] = d->full_state[XX_AB];
  /* Only gives positive answers
  state[X_AKL] = pi - ArcCos((-Power(sl,2) - Power(thl,2) + 
			      Power(-(sl*casl) + thl*cathl,2) + 
			      Power(sl*sasl - thl*sathl,2))/(2.*sl*thl));
  state[X_AKR] = pi - ArcCos((-Power(sl,2) - Power(thl,2) + 
			      Power(-(sl*casr) + thl*cathr,2) + 
			      Power(sl*sasr - thl*sathr,2))/(2.*sl*thl));
  */
  state[X_AKL] = athl - asl;
  state[X_AKR] = -(athr - asr);
  state[X_AH] = d->full_state[XX_ATO];

  state[X_XRD] = d->full_state[XX_XR + N_XX_ACC];
  state[X_ABD] = d->full_state[XX_AB + N_XX_ACC];
  /*
  state[X_AKLD] = (asld - athld)*Csc(asl - athl)*Sqrt(Power(Sin(asl - athl),2));
  state[X_AKRD] = (asrd - athrd)*Csc(asr - athr)*Sqrt(Power(Sin(asr - athr),2));
  */
  state[X_AKLD] = athld - asld;
  state[X_AKRD] = -(athrd - asrd);
  state[X_AHD] = d->full_state[XX_ATO + N_XX_ACC];

#ifdef COMMENT
  // Let's check state_to_full_state()
  akl = state[X_AKL];
  akr = state[X_AKR];
  akld = state[X_AKLD];
  akrd = state[X_AKRD];
  lll2 = sl*sl + thl*thl - 2*sl*thl*cos( pi - akl );
  llr2 = sl*sl + thl*thl - 2*sl*thl*cos( pi - akr );
  errorl = -asld + 
    /*
   abd - (akld*thl*(thl + sl*Cos(akl)))/
     (Power(sl,2) + Power(thl,2) + 2*sl*thl*Cos(akl)) - 
    (sl*thl*(akld*(2*Power(ao,2) - Power(sl,2) - Power(thl,2) + 
            sl*thl*(Cos(akl - pi) + Cos(akr - pi)))*Sin(akl - pi) + 
         akrd*(Power(sl,2) + Power(thl,2) - 2*sl*thl*Cos(akl - pi))*
          Sin(akr - pi)))/
     (2.*ao*Power(Power(sl,2) + Power(thl,2) - 2*sl*thl*Cos(akl - pi),1.5)*
       Sqrt(1 - Power(2*Power(ao,2) + sl*thl*(-Cos(akl - pi) + Cos(akr - pi)),
           2)/(4.*Power(ao,2)*(Power(sl,2) + Power(thl,2) - 
			       2*sl*thl*Cos(akl - pi)))));
    */
    /*
   abd - (akld*thl*(thl + sl*Cos(akl)))/lll2 + 
    (sl*thl*(akld*(-4*Power(ao,2) + lll2 + llr2)*Sin(akl - pi) - 
         2*akrd*lll2*Sin(akr - pi)))/
     (ao*Power(lll2,1.5)*Sqrt(-((16*Power(ao,4) + Power(lll2 - llr2,2) - 
				 8*Power(ao,2)*(lll2 + llr2))/(Power(ao,2)*lll2))));
    */
    abd - (akld*thl*(thl + sl*Cos(akl)))/lll2 + 
    (sl*thl*(akld*(-4*Power(ao,2) + lll2 + llr2)*Sin(akl - pi) - 
	     2*akrd*lll2*Sin(akr - pi)))/
    (lll2*Sqrt(-((16*Power(ao,4) + Power(lll2 - llr2,2) - 
		  8*Power(ao,2)*(lll2 + llr2)))));

  errorr = -asrd + 
    /*
   abd + (akrd*thl*(thl + sl*Cos(akr)))/
     (Power(sl,2) + Power(thl,2) + 2*sl*thl*Cos(akr)) + 
    (sl*thl*(akld*(Power(sl,2) + Power(thl,2) - 2*sl*thl*Cos(akr - pi))*
          Sin(akl - pi) + akrd*(2*Power(ao,2) - Power(sl,2) - Power(thl,2) + 
            sl*thl*(Cos(akl - pi) + Cos(akr - pi)))*Sin(akr - pi)))/
     (2.*ao*Power(Power(sl,2) + Power(thl,2) - 2*sl*thl*Cos(akr - pi),1.5)*
       Sqrt(1 - Power(2*Power(ao,2) + sl*thl*(Cos(akl - pi) - Cos(akr - pi)),
           2)/(4.*Power(ao,2)*(Power(sl,2) + Power(thl,2) - 
			       2*sl*thl*Cos(akr - pi)))));
    */
    /*
   abd + (akrd*thl*(thl + sl*Cos(akr)))/llr2 + 
    (sl*thl*(2*akld*llr2*Sin(akl - pi) - 
         akrd*(-4*Power(ao,2) + lll2 + llr2)*Sin(akr - pi)))/
     (ao*Power(llr2,1.5)*Sqrt(-((16*Power(ao,4) + Power(lll2 - llr2,2) - 
				 8*Power(ao,2)*(lll2 + llr2))/(Power(ao,2)*llr2))));
    */
   abd + (akrd*thl*(thl + sl*Cos(akr)))/llr2 + 
    (sl*thl*(2*akld*llr2*Sin(akl - pi) - 
         akrd*(-4*Power(ao,2) + lll2 + llr2)*Sin(akr - pi)))/
     (llr2*Sqrt(-((16*Power(ao,4) + Power(lll2 - llr2,2) - 
		   8*Power(ao,2)*(lll2 + llr2)))));

  if ( ( ( fabs( errorl/(asld + 1e-2) ) + fabs( errorr/(asrd + 1e-2) ) ) > 1e-5 )
       && ( ( fabs( errorl) > 1e-5 ) || ( fabs( errorr ) > 1e-5 ) ) )
    printf( "asld/asrd errors: %g %g %g %g\n", errorl, errorr, asld, asrd );
#endif
}

/**********************************************************************/

void full_accelerations_to_accelerations( Dynamics *d, double *xd )
{
  double asl, athl, asr, athr;
  double asldd, athldd, asrdd, athrdd;

  asl = d->full_state[XX_ASL];
  athl = d->full_state[XX_ATHL];
  asr = d->full_state[XX_ASR];
  athr = d->full_state[XX_ATHR];

  asldd = d->nr_b[XX_ASL+1];
  athldd = d->nr_b[XX_ATHL+1];
  asrdd = d->nr_b[XX_ASR+1];
  athrdd = d->nr_b[XX_ATHR+1];

  xd[X_XRD] = d->nr_b[XX_XR+1];
  xd[X_ABD] = d->nr_b[XX_AB+1];
  xd[X_AHD] = d->nr_b[XX_ATO+1];

  // CGA: check this, looks like
  // Csc(asl - athl)*Sqrt(Power(Sin(asl - athl),2)) = sign(asl - athl)
  xd[X_AKLD] = (asldd - athldd)*Csc(asl - athl)*Sqrt(Power(Sin(asl - athl),2));
  xd[X_AKRD] = (asldd - athldd)*Csc(asr - athr)*Sqrt(Power(Sin(asr - athr),2));
}

/**********************************************************************/

void state_to_full_state( Dynamics *d, double *state )
{
  int i;
  double xr0, rr, scm, thcm, tocm;
  double ao, sl, thl;
  double ab, akl, akr, ato;
  double xr;
  double abd, akld, akrd;
  double ll2, lr2, asl, asr, athl, athr;
  double state_mag2;
  double offsetl, offsetr;
  double pi = 3.14159265;
  double sab, cab, sasl, casl, sasr, casr, sathl, cathl, sathr, cathr;
  double sato, cato;

  for ( i = 0; i < N_XX; i++ )
    d->full_state[i] = 0;

  // use Mathematica variable names
  xr0 = d->roller_x0;
  rr = d->roller_radius;
  scm = d->shank_cm;
  thcm = d->thigh_cm;
  tocm = d->torso_cm;
  ao = d->ao;
  sl = d->shank_length;
  thl = d->thigh_length;

  // use Mathematica variable names
  xr = d->full_state[XX_XR];
  ab = state[X_AB];
  akl = state[X_AKL];
  akr = state[X_AKR];
  ato = state[X_AH];

  // do the sin/cos calls we can
  sab = sin(ab);
  cab = cos(ab);
  sato = sin(ato);
  cato = cos(ato);

  // use Mathematica variable names
  abd = state[X_ABD];
  akld = state[X_AKLD];
  akrd = state[X_AKRD];

  // CGA: can clean up
  ll2 = sl*sl + thl*thl - 2*sl*thl*cos( pi - akl );
  lr2 = sl*sl + thl*thl - 2*sl*thl*cos( pi - akr );

  offsetl = atan2( thl*sin( akl ), sl + thl*cos( akl ) );
  offsetr = atan2( thl*sin( akr ), sl + thl*cos( akr ) );

  asl = -(pi/2 - ArcCos((lr2 - ll2 - (4*ao*ao))/(2*sqrt(ll2)*2*ao))) - offsetl + ab;
  asr =  (pi/2 - ArcCos((ll2 - lr2 - (4*ao*ao))/(2*sqrt(lr2)*2*ao)) + offsetr) + ab;

  athl = asl + akl;
  athr = asr - akr;

  /*
  printf( "stfs: %g %g %g; %g %g; %g %g; %g %g; %g %g; %g %g\n",
	  ab, akl, akr, ll2, lr2, offsetl, offsetr,
	  ArcCos((lr2 - ll2 - (4*ao*ao))/(2*ll2*2*ao)),
	  ArcCos((ll2 - lr2 - (4*ao*ao))/(2*lr2*2*ao)),
	  asl, asr, athl, athr );
  */

  d->full_state[XX_XR] = state[X_XR];
  d->full_state[XX_AB] = state[X_AB];
  d->full_state[XX_ASL] = asl;
  d->full_state[XX_ATHL] = athl;
  d->full_state[XX_ASR] = asr;
  d->full_state[XX_ATHR] = athr;
  d->full_state[XX_ATO] = state[X_AH];

  /*
  state_mag2 = 0.0;
  for ( i = N_ACC; i < N_X; i++ )
    state_mag2 += state[i]*state[i];
  if ( fabs( state_mag2 ) > 1e-20 )
    {
      fprintf( stderr, "state_to_full_state doesn't handle non-zero velocities yet.\n" );
      exit( -1 );
    }
  */

  // do the rest of the trig calls
  sasl = sin(asl);
  casl = cos(asl);
  sasr = sin(asr);
  casr = cos(asr);
  sathl = sin(athl);
  cathl = cos(athl);
  sathr = sin(athr);
  cathr = cos(athr);

  d->full_state[XX_XR + N_XX_ACC] = state[X_XRD];
  d->full_state[XX_AB + N_XX_ACC] = state[X_ABD];

  d->full_state[XX_ASL + N_XX_ACC] = 
    /*
    abd + (akrd*thl*(thl + sl*Cos(akr)))/
    (Power(sl,2) + Power(thl,2) + 2*sl*thl*Cos(akr)) + 
    (sl*thl*(akld*(Power(sl,2) + Power(thl,2) - 2*sl*thl*Cos(akr - pi))*
	     Sin(akl - pi) + akrd*(2*Power(ao,2) - Power(sl,2) - Power(thl,2) + 
				   sl*thl*(Cos(akl - pi) + Cos(akr - pi)))*Sin(akr - pi)))/
    (2.*ao*Power(Power(sl,2) + Power(thl,2) - 2*sl*thl*Cos(akr - pi),1.5)*
     Sqrt(1 - Power(2*Power(ao,2) + sl*thl*(Cos(akl - pi) - Cos(akr - pi)),
		    2)/(4.*Power(ao,2)*(Power(sl,2) + Power(thl,2) - 
					2*sl*thl*Cos(akr - pi)))));
    */
    abd - (akld*thl*(thl + sl*Cos(akl)))/ll2 + 
    (sl*thl*(akld*(-4*Power(ao,2) + ll2 + lr2)*Sin(akl - pi) - 
	     2*akrd*ll2*Sin(akr - pi)))/
    (ll2*Sqrt(-((16*Power(ao,4) + Power(ll2 - lr2,2) - 
		  8*Power(ao,2)*(ll2 + lr2)))));
  
  d->full_state[XX_ASR + N_XX_ACC] = 
    /*
   abd + (akrd*thl*(thl + sl*Cos(akr)))/
     (Power(sl,2) + Power(thl,2) + 2*sl*thl*Cos(akr)) + 
    (sl*thl*(akld*(Power(sl,2) + Power(thl,2) - 2*sl*thl*Cos(akr - pi))*
          Sin(akl - pi) + akrd*(2*Power(ao,2) - Power(sl,2) - Power(thl,2) + 
            sl*thl*(Cos(akl - pi) + Cos(akr - pi)))*Sin(akr - pi)))/
     (2.*ao*Power(Power(sl,2) + Power(thl,2) - 2*sl*thl*Cos(akr - pi),1.5)*
       Sqrt(1 - Power(2*Power(ao,2) + sl*thl*(Cos(akl - pi) - Cos(akr - pi)),
           2)/(4.*Power(ao,2)*(Power(sl,2) + Power(thl,2) - 
			       2*sl*thl*Cos(akr - pi)))));
    */
   abd + (akrd*thl*(thl + sl*Cos(akr)))/lr2 + 
    (sl*thl*(2*akld*lr2*Sin(akl - pi) - 
         akrd*(-4*Power(ao,2) + ll2 + lr2)*Sin(akr - pi)))/
     (lr2*Sqrt(-((16*Power(ao,4) + Power(ll2 - lr2,2) - 
		   8*Power(ao,2)*(ll2 + lr2)))));

  d->full_state[XX_ATHL + N_XX_ACC] = d->full_state[XX_ASL + N_XX_ACC] + akld;
  d->full_state[XX_ATHR + N_XX_ACC] = d->full_state[XX_ASR + N_XX_ACC] - akrd;
  d->full_state[XX_ATO + N_XX_ACC] = state[X_AHD];

  forward_kinematics( d );

  /*
  printf( "c: %g %g\n", d->xc, d->yc );
  printf( "al: %g %g\n", d->xal, d->yal );
  printf( "ar: %g %g\n", d->xar, d->yar );
  printf( "kl: %g %g\n", d->xkl, d->ykl );
  printf( "kr: %g %g\n", d->xkr, d->ykr );
  printf( "hl: %g %g\n", d->xh, d->yh );
  printf( "hr: %g %g\n",
	  d->xkr - (thl*sathr),
	  d->ykr + thl*cathr );
  printf( "head: %g %g\n", d->xhead, d->yhead );
  printf( "asxd: %g %g\n", d->full_state[XX_ASL + N_XX_ACC], d->full_state[XX_ASR + N_XX_ACC] );
  */

  d->full_state[XX_YR] = d->roller_radius;
  d->full_state[XX_AR] = -d->full_state[XX_XR]/d->roller_radius;
  d->full_state[XX_XB] = d->xc + (ab*rr + xr - xr0)*cab;
  d->full_state[XX_YB] = d->yc + (ab*rr + xr - xr0)*sab;
  d->full_state[XX_XSL] = d->xal - scm*sasl;
  d->full_state[XX_YSL] = d->yal + scm*casl;
  d->full_state[XX_XTHL] = d->xkl - thcm*sathl;
  d->full_state[XX_YTHL] = d->ykl + thcm*cathl;
  d->full_state[XX_XSR] = d->xar - scm*sasr;
  d->full_state[XX_YSR] = d->yar + scm*casr;
  d->full_state[XX_XTHR] = d->xkr - thcm*sathr;
  d->full_state[XX_YTHR] = d->ykr + thcm*cathr;
  d->full_state[XX_XTO] = d->xh - tocm*sato;
  d->full_state[XX_YTO] = d->yh + tocm*cato;

  d->full_state[XX_YR + N_XX_ACC] = 0.0;
  d->full_state[XX_AR + N_XX_ACC] = -d->full_state[XX_XR + N_XX_ACC]/d->roller_radius;
  d->full_state[XX_XB + N_XX_ACC] = d->xcd + (d->wb*rr + d->xrd)*cab - abd*(ab*rr + xr - xr0)*sab;
  d->full_state[XX_YB + N_XX_ACC] = d->ycd + d->wb*(ab*rr + xr - xr0)*cab + (d->wb*rr + d->xrd)*sab;
  d->full_state[XX_XSL + N_XX_ACC] = d->xald - d->wsl*scm*casl;
  d->full_state[XX_YSL + N_XX_ACC] = d->yald - d->wsl*scm*sasl;
  d->full_state[XX_XTHL + N_XX_ACC] = d->xkld - d->wthl*thcm*cathl;
  d->full_state[XX_YTHL + N_XX_ACC] = d->ykld - d->wthl*thcm*sathl;
  d->full_state[XX_XSR + N_XX_ACC] = d->xard - d->wsr*scm*casr;
  d->full_state[XX_YSR + N_XX_ACC] = d->yard - d->wsr*scm*sasr;
  d->full_state[XX_XTHR + N_XX_ACC] = d->xkrd - d->wthr*thcm*cathr;
  d->full_state[XX_YTHR + N_XX_ACC] = d->ykrd - d->wthr*thcm*sathr;
  d->full_state[XX_XTO + N_XX_ACC] = d->xhd - d->wto*tocm*cato;
  d->full_state[XX_YTO + N_XX_ACC] = d->yhd - d->wto*tocm*sato;
}

/**********************************************************************/

int set_state( Dynamics *d, double time, double *state )
{
  int i; 

  d->time = time;
  for ( i = 0; i < N_X; i++ )
    d->state[i] = state[i];

  state_to_full_state( d, state );

  return 1;
}

/**********************************************************************/

int set_full_state( Dynamics *d, double time, double *full_state )
{
  int i; 

  d->time = time;
  for ( i = 0; i < N_XX; i++ )
    d->full_state[i] = full_state[i];

  forward_kinematics( d );

  full_state_to_state( d, d->state );

  return 1;
}

/**********************************************************************/

void accelerations( Dynamics *d, double *action, double *xd, double *xxd )
{
  int i, j, k;
  double xr0, mr, Ir, rr, mb, Ib, scm, sl, ms, Is, thcm, thl, mth, Ith;
  double tocm, tol, mto, Ito;
  double xr, ab, asl, athl, asr, athr, ato;
  double xrd, abd, asld, athld, asrd, athrd, atod;
  double ao;
  double tbsl, tslthl, tthlto, tbsr, tsrthr, tthrto; // torques
  int row;
  double nr_d;
  double new_velocity;
  double sab, cab, sasl, casl, sasr, casr, sathl, cathl, sathr, cathr;
  double sato, cato;
  
  // use Mathematica variable names
  xr0 = d->roller_x0;
  rr = d->roller_radius;
  mr = d->roller_mass;
  Ir = d->roller_I;
  mb = d->board_mass;
  Ib = d->board_I;
  scm = d->shank_cm;
  sl = d->shank_length;
  ms = d->shank_mass;
  Is = d->shank_I;
  thcm = d->thigh_cm;
  thl = d->thigh_length;
  mth = d->thigh_mass;
  Ith = d->thigh_I;
  tocm = d->torso_cm;
  tol = d->torso_length;
  mto = d->torso_mass;
  Ito = d->torso_I;
  ao = d->ao;

  // use Mathematica variable names
  xr = d->full_state[XX_XR];
  ab = d->full_state[XX_AB];
  asl = d->full_state[XX_ASL];
  athl = d->full_state[XX_ATHL];
  asr = d->full_state[XX_ASR];
  athr = d->full_state[XX_ATHR];
  ato = d->full_state[XX_ATO];

  // do all the sin/cos calls
  sab = sin(ab);
  cab = cos(ab);
  sasl = sin(asl);
  casl = cos(asl);
  sasr = sin(asr);
  casr = cos(asr);
  sathl = sin(athl);
  cathl = cos(athl);
  sathr = sin(athr);
  cathr = cos(athr);
  sato = sin(ato);
  cato = cos(ato);

  // use Mathematica variable names
  xrd = d->full_state[XX_XR + N_XX_ACC];
  abd = d->full_state[XX_AB + N_XX_ACC];
  asld = d->full_state[XX_ASL + N_XX_ACC];
  athld = d->full_state[XX_ATHL + N_XX_ACC];
  asrd = d->full_state[XX_ASR + N_XX_ACC];
  athrd = d->full_state[XX_ATHR + N_XX_ACC];
  atod = d->full_state[XX_ATO + N_XX_ACC];

  // use Mathematica variable names
  tbsl = -action[U_AAL];
  tslthl = -action[U_AKL];
  tthlto = -action[U_AHL];
  tbsr = action[U_AAR];
  tsrthr = action[U_AKR];
  tthrto = action[U_AHR];

  // Numerical Recipes format: indices start at 1
  for ( i = 1; i <= N_B; i++ )
    {
      d->nr_b[i] = 0.0;
      for ( j = 1; j <= N_B; j++ )
	d->nr_a[i][j] = 0.0;
    }

  // Solving Ax = b

  row = 1; // Numerical Recipes format: indices start at 1

  // As much as possible the following is raw Mathematica output

  // GROUND CONSTRAINTS
  // 1 xrdd + rr * ardd = 0
  d->nr_a[row][XX_XR+1] = 1.0;
  d->nr_a[row][XX_AR+1] = rr;
  d->nr_b[row] = -( 0.0 );
  row++;

  // 2 yrdd = 0
  d->nr_a[row][XX_YR+1] = 1.0;
  d->nr_b[row] = -( 0.0 );
  row++;


  // ROLLER DYNAMICS
  // 3 mr*xrdd - fgrx + frbx = 0
  d->nr_a[row][XX_XR+1] = mr;
  d->nr_a[row][XX_FGRX+1] = -1.0;
  d->nr_a[row][XX_FRBX+1] = 1.0;
  d->nr_b[row] = -( 0.0 );
  row++;

  // 4 mr*yrdd - fgry + frby = - mr*G
  d->nr_a[row][XX_YR+1] = mr;
  d->nr_a[row][XX_FGRY+1] = -1.0;
  d->nr_a[row][XX_FRBY+1] = 1.0;
  d->nr_b[row] = -( mr*G );
  row++;

  // 5 Ir*ardd - fgrx*rr - frbx*rr*cab - frby*rr*Sin(ab) = 0
  d->nr_a[row][XX_AR+1] = Ir;
  d->nr_a[row][XX_FGRX+1] = -rr;
  d->nr_a[row][XX_FRBX+1] = -rr*cab;
  d->nr_a[row][XX_FRBY+1] = -rr*sab;
  d->nr_b[row] = -( 0.0 );
  row++;

  // ROLLER/BOARD CONSTRAINTS
  // 6 
  d->nr_a[row][XX_XR+1] = 1.0 + cab;
  d->nr_a[row][XX_XB+1] = -1.0;
  d->nr_a[row][XX_AB+1] = (-ab*rr*sab - (xr - xr0)*sab);
  d->nr_b[row] = -( -ab*Power(abd,2)*rr*cab - Power(abd,2)*xr*cab + Power(abd,2)*xr0*cab - Power(abd,2)*rr*sab - 2*abd*xrd*sab );
  row++;

  // 7
  d->nr_a[row][XX_XR+1] = sab;
  d->nr_a[row][XX_YR+1] = 1.0;
  d->nr_a[row][XX_YB+1] = -1.0;
  d->nr_a[row][XX_AB+1] = ab*rr*cab + xr*cab - xr0*cab;
  d->nr_b[row] = -( Power(abd,2)*rr*cab + 2*abd*xrd*cab - ab*Power(abd,2)*rr*sab - Power(abd,2)*xr*sab + Power(abd,2)*xr0*sab );
  row++;


  // BOARD DYNAMICS
  // 8 mb*xbdd - frbx + fbslx + fbsrx = 0;
  d->nr_a[row][XX_XB+1] = mb;
  d->nr_a[row][XX_FRBX+1] = -1.0;
  d->nr_a[row][XX_FBSLX+1] = 1.0;
  d->nr_a[row][XX_FBSRX+1] = 1.0;
  d->nr_b[row] = -( 0.0 );
  row++;

  // 9 mb*ybdd - frby + fbsly + fbsry = -mb*G
  d->nr_a[row][XX_YB+1] = mb;
  d->nr_a[row][XX_FRBY+1] = -1.0;
  d->nr_a[row][XX_FBSLY+1] = 1.0;
  d->nr_a[row][XX_FBSRY+1] = 1.0;
  d->nr_b[row] = -( mb*G );
  row++;

  // 10 Ib*abdd ...
  d->nr_a[row][XX_AB+1] = Ib;
  d->nr_a[row][XX_FRBX+1] = -(ab*rr + xr - xr0)*sab;
  d->nr_a[row][XX_FRBY+1] = (ab*rr + xr - xr0)*cab;
  d->nr_a[row][XX_FBSLX+1] = -ao*sab;
  d->nr_a[row][XX_FBSLY+1] = ao*cab;
  d->nr_a[row][XX_FBSRX+1] = ao*sab;
  d->nr_a[row][XX_FBSRY+1] = -ao*cab;
  d->nr_b[row] = -( tbsl + tbsr );
  row++;

  // BOARD/LEFT SHANK (ANKLE) CONSTRAINTS
  // 11 
  d->nr_a[row][XX_XB+1] = 1.0;
  d->nr_a[row][XX_AB+1] = -ao*sab;
  d->nr_a[row][XX_XSL+1] = -1.0;
  d->nr_a[row][XX_ASL+1] = -scm*casl;
  d->nr_b[row] = -( -Power(abd,2)*ao*cab + Power(asld,2)*scm*sasl );
  row++;

  // 12
  d->nr_a[row][XX_YB+1] = 1.0;
  d->nr_a[row][XX_AB+1] = ao*cab;
  d->nr_a[row][XX_YSL+1] = -1.0;
  d->nr_a[row][XX_ASL+1] = -scm*sasl;
  d->nr_b[row] = -( -Power(abd,2)*ao*sab - Power(asld,2)*scm*casl );
  row++;

  // LEFT SHANK DYNAMICS
  // 13 ms*xsldd = fbslx - fslthlx
  d->nr_a[row][XX_XSL+1] = ms;
  d->nr_a[row][XX_FBSLX+1] = -1.0;
  d->nr_a[row][XX_FSLTHLX+1] = 1.0;
  d->nr_b[row] = -( 0.0 );
  row++;

  // 14 ms*ysldd = fbsly - fslthly - ms*G
  d->nr_a[row][XX_YSL+1] = ms;
  d->nr_a[row][XX_FBSLY+1] = -1.0;
  d->nr_a[row][XX_FSLTHLY+1] = 1.0;
  d->nr_b[row] = -( ms*G );
  row++;

  // 15 Is*asldd ...
  d->nr_a[row][XX_ASL+1] = Is;
  d->nr_a[row][XX_FBSLX+1] = -scm*casl;
  d->nr_a[row][XX_FBSLY+1] = -scm*sasl;
  d->nr_a[row][XX_FSLTHLX+1] = -(sl - scm)*casl;
  d->nr_a[row][XX_FSLTHLY+1] = -(sl - scm)*sasl;
  d->nr_b[row] = tbsl;
  d->nr_b[row] = -tslthl;
  row++;

  // LEFT SHANK/LEFT THIGH (KNEE) CONSTRAINTS
  // 16 
  d->nr_a[row][XX_XSL+1] = 1.0;
  d->nr_a[row][XX_ASL+1] = -(sl - scm)*casl;
  d->nr_a[row][XX_XTHL+1] = -1.0;
  d->nr_a[row][XX_ATHL+1] = -thcm*cathl;
  d->nr_b[row] = -( -Power(asld,2)*scm*sasl + Power(asld,2)*sl*sasl + Power(athld,2)*thcm*sathl );
  row++;

  // 17
  d->nr_a[row][XX_YSL+1] = 1.0;
  d->nr_a[row][XX_ASL+1] = -(sl - scm)*sasl;
  d->nr_a[row][XX_YTHL+1] = -1.0;
  d->nr_a[row][XX_ATHL+1] = -thcm*sathl;
  d->nr_b[row] = -( Power(asld,2)*scm*casl - Power(asld,2)*sl*casl - Power(athld,2)*thcm*cathl );
  row++;

  // LEFT THIGH DYNAMICS
  // 18 mth*xthldd = fslthlx - fthltox
  d->nr_a[row][XX_XTHL+1] = mth;
  d->nr_a[row][XX_FSLTHLX+1] = -1.0;
  d->nr_a[row][XX_FTHLTOX+1] = 1.0;
  d->nr_b[row] = -( 0.0 );
  row++;

  // 19 mth*ythldd = fslthly - fthltoy - mth*G
  d->nr_a[row][XX_YTHL+1] = mth;
  d->nr_a[row][XX_FSLTHLY+1] = -1.0;
  d->nr_a[row][XX_FTHLTOY+1] = 1.0;
  d->nr_b[row] = -( mth*G );
  row++;

  // 20 Ith*athldd ...
  d->nr_a[row][XX_ATHL+1] = Ith;
  d->nr_a[row][XX_FSLTHLX+1] = -thcm*cathl;
  d->nr_a[row][XX_FSLTHLY+1] = -thcm*sathl;
  d->nr_a[row][XX_FTHLTOX+1] = -(thl - thcm)*cathl;
  d->nr_a[row][XX_FTHLTOY+1] = -(thl - thcm)*sathl;
  d->nr_b[row] = tslthl;
  d->nr_b[row] = -tthlto;
  row++;

  // LEFT THIGH/TORSO (HIP) CONSTRAINTS
  // 21 
  d->nr_a[row][XX_XTHL+1] = 1.0;
  d->nr_a[row][XX_ATHL+1] = -(thl - thcm)*cathl;
  d->nr_a[row][XX_XTO+1] = -1.0;
  d->nr_a[row][XX_ATO+1] = -tocm*cato;
  d->nr_b[row] = -( (thl - thcm)*Power(athld,2)*sathl + Power(atod,2)*tocm*sato );
  row++;

  // 22
  d->nr_a[row][XX_YTHL+1] = 1.0;
  d->nr_a[row][XX_ATHL+1] = -(thl - thcm)*sathl;
  d->nr_a[row][XX_YTO+1] = -1.0;
  d->nr_a[row][XX_ATO+1] = -tocm*sato;
  d->nr_b[row] = -( -(thl - thcm)*Power(athld,2)*cathl - Power(atod,2)*tocm*cato );
  row++;

  // BOARD/RIGHT SHANK (ANKLE) CONSTRAINTS
  // 23 
  d->nr_a[row][XX_XB+1] = 1.0;
  d->nr_a[row][XX_AB+1] = ao*sab;
  d->nr_a[row][XX_XSR+1] = -1.0;
  d->nr_a[row][XX_ASR+1] = -scm*casr;
  d->nr_b[row] = -( Power(abd,2)*ao*cab + Power(asrd,2)*scm*sasr );
  row++;

  // 24
  d->nr_a[row][XX_YB+1] = 1.0;
  d->nr_a[row][XX_AB+1] = -ao*cab;
  d->nr_a[row][XX_YSR+1] = -1.0;
  d->nr_a[row][XX_ASR+1] = -scm*sasr;
  d->nr_b[row] = -( Power(abd,2)*ao*sab - Power(asrd,2)*scm*casr );
  row++;

  // RIGHT SHANK DYNAMICS
  // 25 ms*xsrdd = fbsrx - fsrthrx
  d->nr_a[row][XX_XSR+1] = ms;
  d->nr_a[row][XX_FBSRX+1] = -1.0;
  d->nr_a[row][XX_FSRTHRX+1] = 1.0;
  d->nr_b[row] = -( 0.0 );
  row++;

  // 26 ms*ysrdd = fbsry - fsrthry - ms*G
  d->nr_a[row][XX_YSR+1] = ms;
  d->nr_a[row][XX_FBSRY+1] = -1.0;
  d->nr_a[row][XX_FSRTHRY+1] = 1.0;
  d->nr_b[row] = -( ms*G );
  row++;

  // 27 Is*asrdd ...
  d->nr_a[row][XX_ASR+1] = Is;
  d->nr_a[row][XX_FBSRX+1] = -scm*casr;
  d->nr_a[row][XX_FBSRY+1] = -scm*sasr;
  d->nr_a[row][XX_FSRTHRX+1] = -(sl - scm)*casr;
  d->nr_a[row][XX_FSRTHRY+1] = -(sl - scm)*sasr;
  d->nr_b[row] = tbsr;
  d->nr_b[row] = -tsrthr;
  row++;

  // RIGHT SHANK/RIGHT THIGH (KNEE) CONSTRAINTS
  // 28 
  d->nr_a[row][XX_XSR+1] = 1.0;
  d->nr_a[row][XX_ASR+1] = -(sl - scm)*casr;
  d->nr_a[row][XX_XTHR+1] = -1.0;
  d->nr_a[row][XX_ATHR+1] = -thcm*cathr;
  d->nr_b[row] = -( -Power(asrd,2)*scm*sasr + Power(asrd,2)*sl*sasr + Power(athrd,2)*thcm*sathr );
  row++;

  // 29
  d->nr_a[row][XX_YSR+1] = 1.0;
  d->nr_a[row][XX_ASR+1] = -(sl - scm)*sasr;
  d->nr_a[row][XX_YTHR+1] = -1.0;
  d->nr_a[row][XX_ATHR+1] = -thcm*sathr;
  d->nr_b[row] = -( Power(asrd,2)*scm*casr - Power(asrd,2)*sl*casr - Power(athrd,2)*thcm*cathr );
  row++;

  // RIGHT THIGH DYNAMICS
  // 30 mth*xthrdd = fsrthrx - fthrtox
  d->nr_a[row][XX_XTHR+1] = mth;
  d->nr_a[row][XX_FSRTHRX+1] = -1.0;
  d->nr_a[row][XX_FTHRTOX+1] = 1.0;
  d->nr_b[row] = -( 0.0 );
  row++;

  // 31 mth*ythrdd = fsrthry - fthrtoy - mth*G
  d->nr_a[row][XX_YTHR+1] = mth;
  d->nr_a[row][XX_FSRTHRY+1] = -1.0;
  d->nr_a[row][XX_FTHRTOY+1] = 1.0;
  d->nr_b[row] = -( mth*G );
  row++;

  // 32 Ith*athrdd ...
  d->nr_a[row][XX_ATHR+1] = Ith;
  d->nr_a[row][XX_FSRTHRX+1] = -thcm*cathr;
  d->nr_a[row][XX_FSRTHRY+1] = -thcm*sathr;
  d->nr_a[row][XX_FTHRTOX+1] = -(thl - thcm)*cathr;
  d->nr_a[row][XX_FTHRTOY+1] = -(thl - thcm)*sathr;
  d->nr_b[row] = tsrthr;
  d->nr_b[row] = -tthrto;
  row++;

  // RIGHT THIGH/TORSO (HIP) CONSTRAINTS
  // 33 
  d->nr_a[row][XX_XTHR+1] = 1.0;
  d->nr_a[row][XX_ATHR+1] = -(thl - thcm)*cathr;
  d->nr_a[row][XX_XTO+1] = -1.0;
  d->nr_a[row][XX_ATO+1] = -tocm*cato;
  d->nr_b[row] = -( (thl - thcm)*Power(athrd,2)*sathr + Power(atod,2)*tocm*sato );
  row++;

  // 34
  d->nr_a[row][XX_YTHR+1] = 1.0;
  d->nr_a[row][XX_ATHR+1] = -(thl - thcm)*sathr;
  d->nr_a[row][XX_YTO+1] = -1.0;
  d->nr_a[row][XX_ATO+1] = -tocm*sato;
  d->nr_b[row] = -( -(thl - thcm)*Power(athrd,2)*cathr - Power(atod,2)*tocm*cato );
  row++;

  // TORSO DYNAMICS
  // 35 mto*xtoldd = fthltolx
  d->nr_a[row][XX_XTO+1] = mto;
  d->nr_a[row][XX_FTHLTOX+1] = -1.0;
  d->nr_a[row][XX_FTHRTOX+1] = -1.0;
  d->nr_b[row] = -( 0.0 );
  row++;

  // 36 mto*ytoldd = fthltoly - mto*G
  d->nr_a[row][XX_YTO+1] = mto;
  d->nr_a[row][XX_FTHLTOY+1] = -1.0;
  d->nr_a[row][XX_FTHRTOY+1] = -1.0;
  d->nr_b[row] = -( mto*G );
  row++;

  // 37 Ito*atodd ...
  d->nr_a[row][XX_ATO+1] = Ito;
  d->nr_a[row][XX_FTHLTOX+1] = -tocm*cato;
  d->nr_a[row][XX_FTHLTOY+1] = -tocm*sato;
  d->nr_a[row][XX_FTHRTOX+1] = -tocm*cato;
  d->nr_a[row][XX_FTHRTOY+1] = -tocm*sato;
  d->nr_b[row] = tthlto + tthrto;
  row++;

  if ( row != N_B + 1 )
    {
      fprintf( stderr, "Row error: %d != %d\n", row, N_B + 1 );
      exit( -1 );
    }

  /*
  for ( i = 1; i <= N_A; i++ )
    printf( "b[%d]: %g\n", i, d->nr_b[i] );
  fflush( stdout );
  */

  /*
  for ( i = 1; i <= N_A; i++ )
    {
      printf( "%d: ", i );
      for ( j = 1; j <= N_A; j++ )
	printf( "%g ", d->nr_a[i][j] );
      printf( "\n" );
    }
  */

  ludcmp( d->nr_a, N_B, d->nr_indx, &nr_d );
  lubksb( d->nr_a, N_B, d->nr_indx, d->nr_b );

  /*
  for ( i = 1; i <= N_A; i++ )
    printf( "bb[%d]: %g\n", i, d->nr_b[i] );
  fflush( stdout );
  */

  if ( xd != NULL )
    {
      for ( i = 0; i < N_ACC; i++ )
	xd[i] = d->state[i + N_ACC];
      full_accelerations_to_accelerations( d, xd );
    }

  if ( xxd != NULL )
    {
      for ( i = 0; i < N_XX_ACC; i++ )
	{
	  xxd[i] = d->full_state[i + N_XX_ACC];
	  xxd[i + N_XX_ACC] = d->nr_b[i + 1];
	}
    }

  /*
  for ( i = 1; i <= N_X; i++ )
    printf( "xd[%d]: %g\n", i, xd[i] );
  fflush( stdout );
  */

  /*
  printf( "Press return to continue.\n" );
  getchar();
  */

  // printf( "yrdd: %g\n", d->nr_b[XX_YR+1] );
}

/**********************************************************************/
/**********************************************************************/

Dynamics *create_dynamics()
{
  int i;
  Dynamics *result;
  double the_desired_state[N_X];

  result = (Dynamics *) malloc( sizeof( Dynamics ) );
  if ( result == NULL )
    {
      fprintf( stderr, "Can't allocate dynamics.\n" );
      exit( -1 );
    }

  result->seed = 1; /* Set random number generator seed */

  result->time = 0;
  result->time_step = 0.001;

  result->status = NOT_CRASHED;

  result->ao = 0.2;

  result->roller_x0 = 0;
  result->roller_radius = 0.07;
  result->roller_mass = 1.0;
  result->roller_I =
    result->roller_mass*result->roller_radius*result->roller_radius;
  result->board_length = 0.73;
  result->board_mass = 1.0;
  result->board_I =
    result->board_mass*result->board_length*result->board_length/12;
  result->shank_cm = 0.25;
  result->shank_length = 0.5;
  result->shank_mass = 10.0;
  result->shank_I =
    result->shank_mass*result->shank_length*result->shank_length/12;
  result->thigh_cm = 0.25;
  result->thigh_length = 0.5;
  result->thigh_mass = 10.0;
  result->thigh_I =
    result->thigh_mass*result->thigh_length*result->thigh_length/12;
  result->torso_cm = 0.5;
  result->torso_length = 1.0;
  result->torso_mass = 50.0;
  result->torso_I =
    result->torso_mass*result->torso_length*result->torso_length/12;

  result->total_mass = result->roller_mass + result->board_mass
    + 2*result->shank_mass + 2*result->thigh_mass + result->torso_mass;

  // Numerical recipes stuff
  result->nr_a = matrix( 1, N_B, 1, N_B );

  // OBSOLETE STUFF

  result->M = dm( N_ACC, N_ACC );
  result->M_inv = dm( N_ACC, N_ACC );

  // Numerical recipes stuff
  result->nr_tempx1 = matrix( 1, N_ACC, 1, N_ACC );
  result->nr_tempx2 = matrix( 1, N_ACC, 1, N_ACC );

  result->aa = dm( N_X, N_X );
  dm_zero( result->aa, N_X, N_X );
  for ( i = 0; i < N_X; i++ )
    result->aa[i][i] = 1;
  for ( i = 0; i < N_X - 1; i++ )
    result->aa[i][i+1] = result->time_step;

  result->bb = dm( N_X, N_U );
  dm_zero( result->bb, N_X, N_U );
  /*
  for ( i = 0; i < N_X - 1; i++ )
    result->bb[i+1][i] = result->time_step;
  */

  result->qq = dm( N_X, N_X );
  dm_zero( result->qq, N_X, N_X );
  for ( i = 0; i < N_X; i++ )
    result->qq[i][i] = 1;

  result->rr = dm( N_U, N_U );
  dm_zero( result->rr, N_U, N_U );
  for ( i = 0; i < N_U; i++ )
    result->rr[i][i] = 1;

  // END OBSOLETE STUFF

  dv_zero( result->state, N_X );
  dv_zero( result->action, N_U );

  for ( i = 0; i < N_X; i++ )
    the_desired_state[i] = 0.0;
  dv_copy( the_desired_state, result->desired_state, N_X );
  set_state( result, 0.0, the_desired_state );

  return result;
}

/**********************************************************************/

int policy( Dynamics *d, double *state, double *parameters, int index,
	    double *action )
{
  int iu, ix;
  int ind = 0;

  for ( iu = 0; iu < N_U; iu++ )
    {
      action[iu] = 0;
      for ( ix = 0; ix < N_X; ix++ )
	{
	  action[iu]
	    += parameters[ind++]*(state[ix] - d->desired_state[ix]);
	}
    }
}

/**********************************************************************/

int integrate_one_step( Dynamics *d, double *action, double *next_state )
//int integrate_one_step_nonlinear( Dynamics *d, double *action, double *next_state )
{
  int i, j;
  double xd[N_X];
  double xxd[N_XX];
  double new_velocity;
  double h;
  double k1[N_X], k2[N_X], k3[N_X], k4[N_X], x_tmp[N_X], start_state[N_X], state[N_X];
  double xx_k1[N_XX], xx_k2[N_XX], xx_k3[N_XX], xx_k4[N_XX], xx_tmp[N_XX], xx_start_state[N_XX], xx_state[N_XX];

  if ( action == NULL )
    action = d->action;

  dv_copy( d->state, start_state, N_X );
  dv_copy( d->full_state, xx_start_state, N_XX );
  h = d->time_step;

  /*
  // Euler method for accelerations and velocities
  accelerations( d, action, xd );

  for ( i = 0; i < N_X; i++ )
    d->next_state[i] = start_state[i] + h*xd[i];

  d->time += d->time_step;
  // End EULER
  */

  /*
  // Euler method for accelerations, modified Euler for velocities
  accelerations( d, action, xd );

  for ( i = 0; i < N_ACC; i++ )
    {
      new_velocity = d->state[i + N_ACC] + xd[i + N_ACC]*d->time_step;
      d->next_state[i] = d->state[i] + d->time_step*(new_velocity + d->state[i + N_ACC])/2;
      d->next_state[i + N_ACC] = new_velocity;
    }

  d->time += d->time_step;
  // End EULER2
  */

  // 4th order Runge Kutta method for accelerations and velocities

  // Get k1
  accelerations( d, action, xd, xxd );
  dv_scale( xd, h, k1, N_X );
  dv_scale( xxd, h, xx_k1, N_XX );

  // Get k2
  d->time += 0.5*d->time_step;
  dv_scale( k1, 0.5, x_tmp, N_X );
  dv_add( start_state, x_tmp, state, N_X );
  dv_scale( xx_k1, 0.5, xx_tmp, N_XX );
  dv_add( xx_start_state, xx_tmp, xx_state, N_XX );

  // use new full state
  set_full_state( d, d->time, xx_state );
  accelerations( d, action, xd, xxd );

  dv_scale( xd, h, k2, N_X );
  dv_scale( xxd, h, xx_k2, N_XX );

  // Get k3
  dv_scale( k2, 0.5, x_tmp, N_X );
  dv_add( start_state, x_tmp, state, N_X );
  dv_scale( xx_k2, 0.5, xx_tmp, N_XX );
  dv_add( xx_start_state, xx_tmp, xx_state, N_XX );

  // use new full state
  set_full_state( d, d->time, xx_state );
  accelerations( d, action, xd, xxd );

  dv_scale( xd, h, k3, N_X );
  dv_scale( xxd, h, xx_k3, N_XX );

  // Get k4
  d->time += 0.5*d->time_step;
  dv_add( start_state, k3, state, N_X );
  dv_add( xx_start_state, xx_k3, xx_state, N_XX );

  // use new full state
  set_full_state( d, d->time, xx_state );
  accelerations( d, action, xd, xxd );

  dv_scale( xd, h, k4, N_X );
  dv_scale( xxd, h, xx_k4, N_XX );

  dv_add( k2, k3, x_tmp, N_X );
  dv_scale( x_tmp, 2.0, x_tmp, N_X );
  dv_acc( k1, x_tmp, N_X );
  dv_acc( k4, x_tmp, N_X );
  dv_scale( x_tmp, 1.0/6.0, x_tmp, N_X );
  dv_add( start_state, x_tmp, state, N_X );

  dv_add( xx_k2, xx_k3, xx_tmp, N_XX );
  dv_scale( xx_tmp, 2.0, xx_tmp, N_XX );
  dv_acc( xx_k1, xx_tmp, N_XX );
  dv_acc( xx_k4, xx_tmp, N_XX );
  dv_scale( xx_tmp, 1.0/6.0, xx_tmp, N_XX );
  dv_add( xx_start_state, xx_tmp, xx_state, N_XX );

  // End RK4

  // What about RK4 with trapezoidal integration of velocities?

  // use new full state
  set_full_state( d, d->time, xx_state );

  if ( next_state != NULL )
    {
      for( i = 0; i < N_X; i++ )
	next_state[i] = d->state[i];
    }

  if ( 0 )
    {
      printf( "next_state: " );
      for( j = 0; j < N_X; j++ )
	{
	  printf( "%20.15g ", d->state[j] );
	}
      printf( "\n" );
    }

  // make sure up to date accelerations stored with old actions in data file.
  accelerations( d, action, xd, xxd ); 

  // stash action in data structure.
  dv_copy( action, d->action, N_U );

  return 0;
}

/**********************************************************************/

int integrate_one_step_alt( Dynamics *d,
			    double *state,
			    double *action,
			    int index,
			    double *next_state )
{
  set_state( d, 0.0, state );
  integrate_one_step( d, action, next_state );
  return 0;
}

/**********************************************************************/

double one_step_cost( Dynamics *d, double *state, double *action, int index )
{
  double ds[N_X], s2[N_X];
  double cx, cu;
  double u2[N_U];

  dv_subtract( state, d->desired_state, ds, N_X );
  dmv_mult( d->qq, ds, s2, N_X, N_X );
  dv_dot( ds, s2, &cx, N_X );

  dmv_mult( d->rr, action, u2, N_U, N_U );
  dv_dot( action, u2, &cu, N_U );

  return 0.5*(cx + cu);
}

/**********************************************************************/
// return TRUE if crash, FALSE otherwise.

int crash( Dynamics *d )
{
  int i;
  double xr0, rr, xr, yb, ab, sab, cab;

  // use Mathematica variable names
  xr0 = d->roller_x0;
  rr = d->roller_radius;
  xr = d->state[X_XR];
  yb = d->full_state[XX_YB];
  ab = d->state[X_AB];
  sab = sin( ab );
  cab = cos( ab );

  // board end hits ground
  if ( yb + 0.5*d->board_length*sab <= 0.0
       || yb - 0.5*d->board_length*sab <= 0.0 )
    return CRASH_BOARD_HITS_GROUND;

  // board runs off roller: assumed board centered at xr0
  if ( fabs(ab*rr + xr - xr0) > d->board_length/2 )
    return CRASH_BOARD_OFF_ROLLER;
    
  // humanoid too low
  if ( d->yh < 2*d->roller_radius
       || d->yhead < 2*d->roller_radius
       || d->ykl < 2*d->roller_radius
       || d->ykr < 2*d->roller_radius )
    return CRASH_TOO_LOW;

  // check knees bending backwards
  if ( d->state[X_AKL] < 0.0 || d->state[X_AKR] < 0.0 )
    return CRASH_KNEE_BACKWARDS;

  // The contact forces must be positive (or contact is lost). 
  // ground/roller contact force:
  if ( d->nr_b[XX_FGRY+1] < 0.0 )
    return CRASH_GR_FORCE_BAD;
  // roller/board contact force
  if ( cab*d->nr_b[XX_FRBY+1] - sab*d->nr_b[XX_FRBX+1] < 0.0 )
    return CRASH_RB_FORCE_BAD;
  // board/left-foot contact force
  if ( cab*d->nr_b[XX_FBSLY+1] - sab*d->nr_b[XX_FBSLX+1] < 0.0 )
    return CRASH_BSL_FORCE_BAD;
  // board/right-foot contact force
  if ( cab*d->nr_b[XX_FBSLY+1] - sab*d->nr_b[XX_FBSLX+1] < 0.0 )
    return CRASH_BSR_FORCE_BAD;

  // Friction cones should not be violated (no slipping contacts).
  // ground/roller
  if ( fabs( d->nr_b[XX_FGRX+1]/d->nr_b[XX_FGRY+1] ) > 1.0 )
    return CRASH_GR_SLIP;
  // roller/board
  if ( fabs( (sab*d->nr_b[XX_FRBY+1] + cab*d->nr_b[XX_FRBX+1])
	     /(cab*d->nr_b[XX_FRBY+1] - sab*d->nr_b[XX_FRBX+1]) ) > 1.0 )
    return CRASH_RB_SLIP;
  // board/left-foot
  if ( fabs( (sab*d->nr_b[XX_FBSLY+1] + cab*d->nr_b[XX_FBSLX+1])
	     /(cab*d->nr_b[XX_FBSLY+1] - sab*d->nr_b[XX_FBSLX+1]) ) > 1.0 )
    return CRASH_BSL_SLIP;
  // board/right-foot
  if ( fabs( (sab*d->nr_b[XX_FBSRY+1] + cab*d->nr_b[XX_FBSRX+1])
	     /(cab*d->nr_b[XX_FBSRY+1] - sab*d->nr_b[XX_FBSRX+1]) ) > 1.0 )
    return CRASH_BSR_SLIP;

  // The maximum actuator torques can not be exceeded.
  for ( i = 0; i < N_U; i++ )
    if ( fabs( d->action[i] ) > U_MAX )
      return CRASH_TORQUE_TOO_BIG;

  return NOT_CRASHED; // 0
}

/**********************************************************************/

const char *crash_messages[] =
  {
    "NOT CRASHED",
    "CRASH_BOARD_HITS_GROUND",
    "CRASH_BOARD_OFF_ROLLER",
    "CRASH_TOO_LOW",
    "CRASH_GR_FORCE_BAD",
    "CRASH_RB_FORCE_BAD",
    "CRASH_BSL_FORCE_BAD",
    "CRASH_BSR_FORCE_BAD",
    "CRASH_GR_SLIP",
    "CRASH_RB_SLIP",
    "CRASH_BSL_SLIP",
    "CRASH_BSR_SLIP",
    "CRASH_KNEE_BACKWARDS",
    "CRASH_TORQUE_TOO_BIG"
  };

void print_crash_message( Dynamics *d, int error_number )
{
  if ( d->status == error_number )
    return;

  if ( error_number <= FATAL_CRASH && error_number > 0 )
    printf( "FATAL: %g: %s\n", d->time, crash_messages[ error_number ] );
  else
    printf( "%g: %s\n", d->time, crash_messages[ error_number ] );
}

/**********************************************************************/


