/* Some indices and sizes */
// state indices:
#define X_XR 0   // location of roller
#define X_AB 1   // board angle
#define X_AKL 2  // two relative knee angles (left and right)
#define X_AKR 3
#define X_AH 4   // absolute (with respect to vertical) torso angle
#define X_XRD 5  // derivatives of above
#define X_ABD 6
#define X_AKLD 7
#define X_AKRD 8
#define X_AHD 9
#define N_X 10
#define N_ACC (N_X/2)

// command indices
#define U_AAL 0  // left ankle torque
#define U_AKL 1  // left knee torque
#define U_AHL 2  // left hip torque
#define U_AAR 3  // right ankle torque
#define U_AKR 4  // right knee torque
#define U_AHR 5  // right hip torque
#define N_U 6

// In the full state each link has its own state: x, y, angle, and derivatives
// acceleration/constraint-force vector (b) indices;
#define XX_XR 0        // roller state
#define XX_YR 1
#define XX_AR 2
#define XX_XB 3        // board state
#define XX_YB 4
#define XX_AB 5
#define XX_XSL 6       // left shank state
#define XX_YSL 7
#define XX_ASL 8
#define XX_XTHL 9      // left thigh state
#define XX_YTHL 10
#define XX_ATHL 11
#define XX_XSR 12      // right shank state
#define XX_YSR 13
#define XX_ASR 14
#define XX_XTHR 15     // right thigh state
#define XX_YTHR 16
#define XX_ATHR 17
#define XX_XTO 18      // torso state
#define XX_YTO 19
#define XX_ATO 20
#define XX_FGRX 21     // ground/roller force
#define XX_FGRY 22
#define XX_FRBX 23     // roller/board force
#define XX_FRBY 24
#define XX_FBSLX 25    // board/left-shank force
#define XX_FBSLY 26
#define XX_FSLTHLX 27  // left-shank/left-thigh force
#define XX_FSLTHLY 28
#define XX_FTHLTOX 29  // left-thigh/torso force
#define XX_FTHLTOY 30
#define XX_FBSRX 31    // board/right-shank force
#define XX_FBSRY 32
#define XX_FSRTHRX 33  // right-shank/right-thigh force
#define XX_FSRTHRY 34
#define XX_FTHRTOX 35  // right-thigh/torso force
#define XX_FTHRTOY 36
#define N_B 37         // number of elements of this vector
#define N_XX_ACC 21    // Number of accelerations above
#define N_XX (2*N_XX_ACC) // number of elements in full state vector

// maximum actuator commands (Nm)
#define U_MAX 100.0

// crash conditions
#define NOT_CRASHED 0
#define CRASH_BOARD_HITS_GROUND 1
#define CRASH_BOARD_OFF_ROLLER 2
#define CRASH_TOO_LOW 3
#define CRASH_GR_FORCE_BAD 4
#define CRASH_RB_FORCE_BAD 5
#define CRASH_BSL_FORCE_BAD 6
#define CRASH_BSR_FORCE_BAD 7
#define CRASH_GR_SLIP 8
#define CRASH_RB_SLIP 9 
#define CRASH_BSL_SLIP 10
#define CRASH_BSR_SLIP 11
#define CRASH_KNEE_BACKWARDS 12
#define CRASH_TORQUE_TOO_BIG 13
#define FATAL_CRASH 11

/* Treat the contents of this structure as hidden */
typedef struct dynamics_struct
{
  double time;
  double time_step;
  double state[N_X];
  double action[N_U];
  double desired_state[N_X];
  unsigned seed; /* random number generator seed for rand_r() */

  // the full (redundant) state is maintained along with the minimal state.
  double full_state[N_XX];

  int status; // current status of simulation

  // parameters
  double roller_x0;
  double roller_radius;
  double roller_mass;
  double roller_I;
  double board_length;
  double board_mass;
  double board_I;
  double shank_length;
  double shank_cm;
  double shank_mass;
  double shank_I;
  double thigh_length;
  double thigh_cm;
  double thigh_mass;
  double thigh_I;
  double torso_length;
  double torso_cm;
  double torso_mass;
  double torso_I;
  double total_mass;
  double ao; // foot spacing

  // forward kinematics: this stuff duplicates what is in full state,
  // but will still be here when Lagrangian approach is used.
  double xr; // roller COM
  double yr;
  double ar;
  double xc; // roller/board contact point
  double yc;
  double xb; // board COM
  double yb;
  double ab;
  double xal; // left ankle location
  double yal;
  double xsl; // left shank COM
  double ysl;
  double asl;
  double xkl; // left knee location
  double ykl;
  double xthl; // left thigh COM
  double ythl;
  double athl;
  double xar; // right ankle location
  double yar;
  double xsr; // right shank COM
  double ysr;
  double asr;
  double xkr; // right knee location
  double ykr;
  double xthr; // right thigh COM
  double ythr;
  double athr;
  double xh; // hip location
  double yh;
  double xto; // torso COM
  double yto;
  double ato;
  double xhead; // head location
  double yhead;

  double xrd; // velocities of above
  double yrd;
  double xcd;
  double ycd;
  double xbd;
  double ybd;
  double xald;
  double yald;
  double xsld;
  double ysld;
  double xkld;
  double ykld;
  double xthld;
  double ythld;
  double xard;
  double yard;
  double xsrd;
  double ysrd;
  double xkrd;
  double ykrd;
  double xthrd;
  double ythrd;
  double xhd;
  double yhd;
  double xtod;
  double ytod;
  double xheadd;
  double yheadd;

  // angular velocities
  double wr; // roller angular velocity
  double wb; // board " "
  double wsl; // left shank " "
  double wthl; // left thigh " "
  double wsr; // right shank " "
  double wthr; // right thigh " "
  double wto; // torso " "

  // COM/COP info
  double xcom; // x of COM location
  double ycom; // y of COM location
  double dx;   // COM - COP (xcom - xr)
  double xcomd; // derivatives of above
  double ycomd;
  double dxd;

  // Energies for debugging
  double rke; // roller kinetic energy
  double rpe; // roller potential energy
  double bke; // board
  double bpe;
  double slke; // left shank
  double slpe;
  double thlke; // left thigh
  double thlpe;
  double srke; // right shank
  double srpe;
  double thrke; // right thigh
  double thrpe;
  double toke; // torso
  double tope;
  double ke; // total kinetic energy
  double pe; // total potential energy
  double energy; // total energy

  // big matrix and vector used in acceleration calculation
  // Numerical Recipes format: indices start with 1
  double **nr_a;
  double nr_b[N_B+1];
  int nr_indx[N_B+1];

  // OBSOLETE
  // Make some matrices available
  double **aa; // A matrix
  double **bb; // B matrix
  double **qq; // Q matrix
  double **rr; // R matrix

  double **M;
  double **M_inv;
  double fd[N_ACC];

  double nr_d[N_ACC+1];
  double nr_e[N_ACC+1];
  double **nr_tempx1;
  double **nr_tempx2;
  // END OBSOLETE

} Dynamics;
  
/* The dynamics API */
int init_dynamics();
Dynamics *create_dynamics();
int set_state( Dynamics *d, double time, double *state );
int integrate_one_step( Dynamics *d, double *action, double *next_state );
int integrate_one_step_alt( Dynamics *d, double *state, double *action,
			    int index, double *next_state );
double one_step_cost( Dynamics *d, double *state, double *action, int index );
int policy( Dynamics *d, double *state, double *parameters, int index,
	    double *action );

/* It is handy to have this global */
extern double the_desired_state[];
