/*****************************************************************************/
// Includes

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#ifdef WIN32
#include <windows.h>
#else
#include <time.h>
#endif

#include "useful/useful.h"
#include "dynamics.h"

/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/
/************************************************************************/
// Used for timing how long simulation takes.
// Comment out if necessary, not essential.

#ifdef WIN32
double get_time()
{
  return (double) timeGetTime()/1000.0;
}
#else
double get_time()
{
  struct timespec the_tp;

  clock_gettime( CLOCK_MONOTONIC, &the_tp );
  return ((double) (the_tp.tv_sec)) + 1.0e-9*the_tp.tv_nsec;
}
#endif

/*****************************************************************************/
// Stuff used in policy

double my_goal_state[N_X];
double my_feedforward[N_U];

#define KNEE_GOAL 0.2
#define KNEE_TORQUE 25.61048429046979

/*****************************************************************************/

void init_my_policy( Dynamics *d )
{
  int i;

  for( i = 0; i < N_X; i++ )
    my_goal_state[i] = d->desired_state[i];
  my_goal_state[X_AKL] = KNEE_GOAL;
  my_goal_state[X_AKR] = KNEE_GOAL;
  for( i = 0; i < N_U; i++ )
    my_feedforward[i] = 0.0;
  my_feedforward[U_AKL] = KNEE_TORQUE;
  my_feedforward[U_AKR] = KNEE_TORQUE;
}

/*****************************************************************************/
// gain matrix

double k[N_U][N_X] = {
{
   -683.916735086208,      305.87992879971,    -15.5393980515441,     15.3616091951567,     163.808366496605,    -255.745854766121,      123.45864205851,    -15.2436261785126,     15.2433450054126,      63.451933336971, 
},
{
    365.444871303113,    -16.0799532562329,    -217.114858444661,     45.7081205049211,    -241.381727097212,     173.999028067551,    -77.8328706196931,    -2.15751274015748,    -10.3319810574824,    -67.8323036461952, 
},
{
   -42.3418237847979,     -37.059967810827,     141.984107395111,     36.4803508134549,    -174.849846355655,     65.3439759297828,    -34.3894006304088,     12.6396671006739,    0.396305581675073,    -39.0693375985248, 
},
{
    683.916735086208,     -305.87992879971,     15.5393980515441,    -15.3616091951567,    -163.808366496605,     255.745854766121,     -123.45864205851,     15.2436261785126,    -15.2433450054126,     -63.451933336971, 
},
{
   -365.547255986593,     15.9478548120763,     46.0080393998881,    -217.133304587736,     241.693783240843,    -174.117882879534,     77.8831569044641,    -10.3544279560858,    -2.18054322120882,     67.9054174669669, 
},
{
    42.4485076841971,     37.1976222606026,     36.3162647120764,     142.151771244474,     174.524672714052,    -65.2201264652793,     34.3370012890295,    0.398291727618497,     12.6422616287753,     38.9931505752792, 
}};

/*****************************************************************************/

void my_policy( Dynamics *d )
{
  int i, j;

  for( i = 0; i < N_U; i++ )
    {
      d->action[i] = 0;
    }

  for ( i = 0; i < N_U; i++ )
    {
      d->action[i] = my_feedforward[i];
      for ( j = 0; j < N_X; j++ )
	{
	  d->action[i] -= k[i][j]*(d->state[j] - my_goal_state[j]);
	}
      if ( d->action[i] > U_MAX )
	d->action[i] = U_MAX;
      if ( d->action[i] < -U_MAX )
	d->action[i] = -U_MAX;
    }
}

/*****************************************************************************/
/*****************************************************************************/

int main( int argc, char **argv )
{
  int i;
  Dynamics *d;
  double last_time = 0;
  double the_time;
  double initial_state[N_X];
  int status;

  last_time = get_time();

  init_dynamics();
  d = create_dynamics();

  init_data( d );

  init_my_policy( d );

  for( i = 0; i < N_X; i++ )
    {
      initial_state[i] = d->desired_state[i];
    }
  initial_state[X_AKL] = 0.2;
  initial_state[X_AKR] = 0.2;

  // test LQR controller
  // initial_state[X_XR] += 0.04;  // 0.05 unstable
  // initial_state[X_AB] += 0.05;  // 0.06 unstable
  // initial_state[X_AKL] += 0.2;  // 0.2 okay, limited by steady state knee bend
  // initial_state[X_AKR] += 0.2; // 0.2 okay, limited by steady state knee bend
  initial_state[X_AH] += 0.1; // 0.08 unstable

  set_state( d, 0.0, initial_state );

  // need to initialize actions to correctly compute forces for computing policy
  for( i = 0; i < N_U; i++ )
    {
      d->action[i] = 0;
    }
  d->action[U_AKL] = KNEE_TORQUE;
  d->action[U_AKR] = KNEE_TORQUE;
  accelerations( d, d->action, NULL, NULL );
  /* used for finding steady state commands to get zero acceleration
  printf( "%g %g %g %g\n",
	  d->nr_b[XX_ASL+1], d->nr_b[XX_ATHL+1], 
	  d->nr_b[XX_ASR+1], d->nr_b[XX_ATHR+1] );
  */

  // 5000 is 5 seconds at 1kHz sampling rate
  for ( i = 0; i < 10000; i++ )
    {
      my_policy( d );
      save_data( d );
      status = crash( d );
      print_crash_message( d, status );
      d->status = status;
      if ( status != NOT_CRASHED && status <= FATAL_CRASH )
	break;
      integrate_one_step( d, NULL, NULL );
    }

  write_the_mrdplot_file( d );

  the_time = get_time();
  printf( "%g seconds to simulate %g seconds.\n", the_time - last_time,
	  d->time );

  return 0;
}

/*****************************************************************************/
